/*
** Module   : ChuiWebSimulator.java
** Abstract : A WebSocket Chui Terminal simulator.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20131113 Created initial version.
** 002 MAG 20140114 Implements WebSockets protocol and communication.
** 003 MAG 20140204 Send text attributes via WebSockets. Add a Watchdog Timer 
**                  guardian and quit command.                   
** 004 MAG 20140211 Implements COLOR.
** 005 CA  20140206 Web socket timeout can't be read from the jetty threads (as these don't have a
**                  P2J context, so it must be passed at construction.
** 006 MAG 20140217 Refactoring web client redirection flaw.
** 007 MAG 20140219 Performances improvements.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;
import java.util.regex.*;

import org.eclipse.jetty.websocket.api.*;
import org.eclipse.jetty.websocket.api.annotations.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.util.*;

/**
 * Chui WebSocket terminal simulator.
 * Implements the WebSockets protocol to communicate with the JS terminal simulator.
 * On socket close terminate the client process.
 */
@SuppressWarnings("serial")
@WebSocket(maxIdleTime = 900000, maxTextMessageSize = 4096)
public class ChuiWebSimulator
extends ChuiSimulator
{  
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ChuiWebSimulator.class.getName()); 
   
   /** Color palette */
   private static final String COLOR_PALETTE = "{\"c\":0, \"p\":%s}"; 
   
   /**  Object representing the other end of this conversation. */
   private RemoteEndpoint remote;  
   
   /** Keyboard input queue */
   private Deque<Integer> keyBuffer = new ConcurrentLinkedDeque<>();
   
   /** Keep cursor status */
   private boolean cursorVisible = true;
   
   /** Web socket timeout. */
   private final long timeout;
   
   /** Watchdog time out value */
   private final long watchdogTimeout;

   /** Watchdog instance */
   private WatchdogTimer watchdog;
   
   /** Color palette JSON message */
   private String paletteMessage;
   
   /** Holds the last screen messages sent to JS code*/
   private volatile Map<Integer, ByteBuffer> lineCache = new ConcurrentHashMap<>();
   
   /** Push messages worker thread */
   private PushMessagesWorker pushWorker;
   
   /**
    * Create a new chui web simulator.
    * 
    * @param    timeout
    *           Web socket timeout.
    * @param    watchdogTimeout
    *           Watchdog thread timeout.
    */
   public ChuiWebSimulator(long timeout, long watchdogTimeout)
   {
      // build palette message
      paletteMessage = buildPaletteMessage();
      this.timeout = timeout;
      this.watchdogTimeout = watchdogTimeout;
      
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
   }
   
   /**
    * This method is called when the remote peer open a web socket connection.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    *
    */
   @OnWebSocketConnect
   public void onConnect(Session session)
   {
      // Stop watchdog timer
      stopWatchdog();
      // Get RemoteEndpoint reference.
      remote = session.getRemote();
      if (timeout > 0)
      {
         session.setIdleTimeout(timeout);
      }
      pushWorker = new PushMessagesWorker(remote);
      // synchronize tty
      initRemoteTerminal();
   }

   /**
    * This method is called when a text message is received.
    *
    * @param   message
    *          The text message.
    */
   @OnWebSocketMessage
   public void onMessage(String message) 
   {
   }

   /**
    * This method is called when a binary message is received.
    *
    * @param   message
    *          Message content as an byte array.
    * @param   offset
    *          Message offset..
    * @param   length
    *          Message size.
    */
   @OnWebSocketMessage
   public void onMessage(byte[] message, int offset, int length)
   {
      if (length == 3 && message[offset] == 0x01)
      {
         int keystroke = (message[offset+1] << 8) & 0xFF00;
         keystroke |= (message[offset+2] & 0x00FF); 
         keyBuffer.offerLast(keystroke);         
      }
   }

   /**
    * This method is called when the connection is closed.
    *
    * @param   statusCode
    *          The status integer code.
    * @param   reason
    *          A string representing the reason.
    */
   @OnWebSocketClose
   public void onClose(int statusCode, String reason)
   {
      pushWorker.stopPushing();
      remote = null;
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
      
      LOG.logp(Level.INFO,
               "ChuiWebSimulator.onClose",
               "",
               LogHelper.generate("Connection closed status=%d reason=%s", statusCode, reason));
   }
   
   /**
    * This method is called on communication errors.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    * @param   error
    *          A <code>Throwable</code> instance representing the error condition.
    */
   @OnWebSocketError
   public void onError(Session session, Throwable error)
   {
      session.close();
      remote = null;
      
      LOG.logp(Level.SEVERE,
               "ChuiWebSimulator.onError",
               "",
               LogHelper.generate("WebSocket communication error!"),
               error);      
   }

   /**
    * Start watchdog timer.
    * 
    * @param   watchdogTimeout
    *          Watchdog timeout value.
    */
   private void startWatchdog(long watchdogTimeout)
   {
      synchronized (keyBuffer)
      {
         if (watchdog == null)
         {
            watchdog = new WatchdogTimer(watchdogTimeout);
         }
      }      
   }

   /**
    * Stop watchdog timer.
    */
   private void stopWatchdog()
   {
      synchronized (keyBuffer)
      {
         if (watchdog != null)
         {
            watchdog.kill();
            // wait for thread exit
            try
            {
               watchdog.join(WatchdogTimer.WAIT_TO_DIE);
            }
            catch (InterruptedException e)
            {
               LOG.logp(Level.INFO,
                        "ChuiWebSimulator.stopWatchdog",
                        "",
                        LogHelper.generate("Watchdog interrupted!"),
                        e);      
            }
            finally
            {
               watchdog = null;
            }
         }
      }      
   }
   
   /**
    * Send a text message.
    * 
    * @param   message
    *          Text message.
    */   
   public void sendTextMessage(String message) 
   {
      if (pushWorker != null)
      {
         pushWorker.pushMessage(message);
      }
   }

   /**
    * Send a binary message.
    * 
    * @param   message
    *          Binary message.
    */   
   public void sendBinaryMessage(byte[] message) 
   {
      if (pushWorker != null)
      {
         pushWorker.pushMessage(ByteBuffer.wrap(message));
      }
   }

   /**
    * Send a binary message.
    * 
    * @param   message
    *          Binary message.
    */   
   public void sendBinaryMessage(ByteBuffer message) 
   {
      if (pushWorker != null)
      {
         pushWorker.pushMessage(message);
      }
   }
   
   /**
    * Whenever the main page is loaded on browser synchronize virtual terminal state.
    */
   private void initRemoteTerminal()
   {
      synchronized (lineCache)
      {
         // clear cache
         lineCache.clear();
         // send color palette
         colorPalette();
         // synchronize
         triggerRepaint(true);
         setCursorStatus(cursorVisible);
         if (cursorVisible)
         {
            setCursor(getCursorX(), getCursorY());
         }
      }
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   public int readKey()
   {
      return keyBuffer.isEmpty() ? -1 : keyBuffer.pollFirst();
   }
   
   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0. Send INIT message down to JS client.
    */   
   @Override
   public void clear()
   {
      super.clear();
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(0x80);
      sendBinaryMessage(message.toByteArray());
   }

   /**
    * The beep() should be implemented by sending a command on the remote end point.
    * The command is part of the communication protocol.
    *
    */
   public void beep()
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(0x84);
      sendBinaryMessage(message.toByteArray());         
   }
   
   /**
    *  Set the cursor visibility. Send SET_CURSOR_ON or SET_CURSOR_OFF message down to JS client.
    *
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequent repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visibility state.
    */   
   @Override
   public boolean setCursorStatus(boolean on) 
   throws IllegalStateException
   {
      cursorVisible = on;
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(0x83);
      message.write(on ? 1 : 0);
      sendBinaryMessage(message.toByteArray());         
      // call super implementation
      return super.setCursorStatus(on);
   }

   /**
    * Position the cursor. Send SET_CURSOR_POS message down to JS client.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean setCursor(int x, int y)
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(0x82);
      message.write(y);
      message.write(x);
      sendBinaryMessage(message.toByteArray());         
      // call super implementation
      return super.setCursor(x, y);
   }

   /**
    * Should be implemented by sending the screen status to the remote end point as a refresh
    * screen operation. The command is part of the communication protocol.
    * The colors and attributes messages are cached and performances are significantly improved.   
    */
   @Override
   public void triggerRepaint(boolean full)
   {
      TerminalOptions options = getTerminalOptions();
      int rows = options.rows;
      int columns = options.columns;
      
      Cell[][] data = getScreenData();
            
      for (int row = 0; row < rows; row++)
      {         
         // build screen line
         ByteArrayOutputStream line = new ByteArrayOutputStream();
         line.write(0x81); // command
         line.write(row);  // line number
       
         for (int col = 0; col < columns; col++)
         {
            Cell cell = data[row][col];
            Color cellColor = cell.color();
            // character
            char c = cell.character();
            line.write((c & 0xFF00) >> 8);
            line.write(c & 0x00FF);
            // attributes
            line.write(cellColor == null ? 0 : cellColor.attribute());
            // FG/BG colors
            int color = cellColor == null ? 0 : cellColor.color();
            if (color == 0)
            {
               line.write(0);               
            }
            else
            {
               int bg = color & 0x07;
               int fg = (color >>> 12) & 0x70;
               line.write(fg | bg);
            }
         }
         // send message
         ByteBuffer payload = ByteBuffer.wrap(line.toByteArray());
         sendCachedMessage(row, payload);
      }            
   }
   
   /**
    * Line messages are cached. Whenever a new line is about to send down to JS client
    * the line cache is checked. If no data for the specify row is found or cached data 
    * is different than the current row data the payload is send down to JS client and
    * the payload is cached.  
    *  
    * @param   row
    *          Row number.
    * @param   payload
    *          Payload data.
    */
   private void sendCachedMessage(int row, ByteBuffer payload) 
   {      
      synchronized (lineCache)
      {         
         ByteBuffer cache = lineCache.get(row);
         if (cache == null || !cache.equals(payload))
         {  
            lineCache.put(row, payload);
            sendBinaryMessage(payload);                  
         }
      }
   }
   
   /**
    * Remote application is about to end. Notify the JS client about that.
    * The client should be redirected to referrer address.
    */
   public void quit()
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(0x85);
      sendBinaryMessage(message.toByteArray());         
   }
         
   /**
    * Build color palette message. This message is the same
    * over the user session so it's build and store on constructor.
    * The message is send whenever a WebSocket connection is open.
    *
    * @return  A string representing the JSON format of color palette.
    */
   private String buildPaletteMessage()
   {
      // build palette
      StringBuilder palette = new StringBuilder("[");      
      for (ColorPalette colorPalette : ColorPalette.values())
      {
         if (palette.length() != 1)
         {
            palette.append(",");
         }
         palette.append("\"").append(colorPalette.asWebColor()).append("\"");
      }      
      palette.append("]");      
      // JSON format
      return String.format(COLOR_PALETTE, palette.toString());
   }
   
   /**
    * Send color palette down to JS client.
    * This is done whenever a WebSocket connection is open.
    */
   private void colorPalette()
   {      
      // send palette
      sendTextMessage(paletteMessage);
   }
}